package com.syntevo.plugin.jira.commit.messagesource;

import java.io.*;

import org.tmatesoft.svn.core.*;
import org.tmatesoft.svn.core.internal.wc2.compat.*;
import org.tmatesoft.svn.core.wc.*;
import org.tmatesoft.svn.core.wc2.*;

import com.syntevo.openapi.gui.*;
import com.syntevo.openapi.smartsvn.command.*;
import com.syntevo.openapi.smartsvn.command.commit.*;
import com.syntevo.openapi.smartsvn.file.*;
import com.syntevo.openapi.smartsvn.settings.*;
import com.syntevo.plugin.jira.commit.*;

/**
 * @author syntevo GmbH
 */
public final class JiraCommitMessageSourceFactory implements ICommitMessageSourceFactory {

	// Constants ==============================================================

	private static final String BUGTRAQ_URL = "bugtraq:url";

	// Fields =================================================================

	private final GuiComponentFactory componentFactory;
	private final SettingsServices settingsServices;
	private final GuiSpacings spacings;

	// Setup ==================================================================

	public JiraCommitMessageSourceFactory(GuiComponentFactory componentFactory, SettingsServices settingsServices, GuiSpacings spacings) {
		this.componentFactory = componentFactory;
		this.settingsServices = settingsServices;
		this.spacings = spacings;
	}

	// Implemented ============================================================

	public ICommitMessageSource createCommitMessageSource(CommitMessageSourceServices sourceServices) {
		final Thread currentThread = Thread.currentThread();
		final ClassLoader originalContextClassLoader = currentThread.getContextClassLoader();
		currentThread.setContextClassLoader(getClass().getClassLoader());
		try {
			final Path commitRootPath = sourceServices.getCommitRootPath();
			if (commitRootPath == null) {
				return null;
			}
			
			final String bugtraqURL = findBugtraqURL(sourceServices.getClients(), commitRootPath);
			if (bugtraqURL == null) {
				return null;
			}

			final JiraUrl url = JiraUrl.parseFromBugtraqUrls(bugtraqURL);
			if (url == null) {
				return null;
			}
			
			final String baseUrl = url.getBaseUrl();
			if (baseUrl == null) {
				return null;
			}

			return new JiraCommitMessageSource(baseUrl, settingsServices, url.getProjectKey(), componentFactory, spacings, sourceServices.getDialogDisplayer());
		}
		finally {
			currentThread.setContextClassLoader(originalContextClassLoader);
		}
	}

	// Utils ==================================================================

	private static String findBugtraqURL(SvnClients clients, Path path) {
		try {
			final SvnOperationFactory operationsFactory = clients.createOperationsFactory();
			for (File directory = path.getFile(); directory != null; directory = directory.getParentFile()) {
				try {
					final SVNPropertyData[] data = new SVNPropertyData[1];
					final SvnGetProperties getProperties = operationsFactory.createGetProperties();
					getProperties.setSingleTarget(SvnTarget.fromFile(directory, SVNRevision.WORKING));
					getProperties.setRevision(SVNRevision.WORKING);
					getProperties.setDepth(SVNDepth.EMPTY);
					getProperties.setReceiver(new ISvnObjectReceiver<SVNProperties>() {
					    public void receive(SvnTarget target, SVNProperties object) throws SVNException {
					        if (object.containsName(BUGTRAQ_URL)) {
					            data[0] = new SVNPropertyData(BUGTRAQ_URL, object.getSVNPropertyValue(BUGTRAQ_URL), operationsFactory.getOptions());
					        }
					    }
					});
					getProperties.run();
					final SVNPropertyData property = data[0];
					if (property != null && property.getValue() != null) {
						final String url = SVNPropertyValue.getPropertyAsString(property.getValue());
						if (!url.startsWith("/")) {
							return url;
						}

						final SVNInfo[] infoContainer = new SVNInfo[1];
						final SvnGetInfo getInfo = operationsFactory.createGetInfo();
						getInfo.setSingleTarget(SvnTarget.fromFile(directory));
						getInfo.setRevision(SVNRevision.WORKING);
						getInfo.setDepth(SVNDepth.EMPTY);
						getInfo.setReceiver(new ISvnObjectReceiver<SvnInfo>() {
							public void receive(SvnTarget target, SvnInfo object) throws SVNException {
								if (infoContainer[0] == null) {
									infoContainer[0] = SvnCodec.info(object);
								}
							}
						});
						getInfo.run();

						final SVNURL rootURL = infoContainer[0].getRepositoryRootURL();
						return getFullURL(url, rootURL);
					}

					if (SVNWCUtil.isWorkingCopyRoot(directory)) {
						return null;
					}
				}
				catch (SVNException ex) {
					// Ignore
				}
			}
		}
		catch (SVNException ex) {
			// Ignore
		}

		return null;
	}

	private static String getFullURL(String shortURL, SVNURL repositoryRoot) {
		if (shortURL.startsWith("/")) {
			final StringBuffer fullURL = new StringBuffer();
			fullURL.append(repositoryRoot.getProtocol());
			fullURL.append("://");
			fullURL.append(repositoryRoot.getHost());
			if (shortURL.startsWith("/:")) {
				fullURL.append(shortURL.substring(1));
			}
			else {
				if (repositoryRoot.hasPort()) {
					fullURL.append(":");
					fullURL.append(repositoryRoot.getPort());
				}
				else {
					fullURL.append(shortURL);
				}
			}

			return fullURL.toString();
		}

		if (shortURL.startsWith("^/")) {
			shortURL = shortURL.substring(2);
			while (shortURL.startsWith("../")) {
				shortURL = shortURL.substring(3);
				try {
					repositoryRoot = repositoryRoot.removePathTail();
				}
				catch (SVNException ex) {
					// We have been at the root
				}
			}

			if (shortURL.startsWith("/")) {
				shortURL = shortURL.substring(1);
			}

			String fullURL = repositoryRoot.toString();
			if (fullURL.endsWith("/")) {
				fullURL = fullURL.substring(0, fullURL.length() - 1);
			}
			return fullURL + "/" + shortURL;
		}

		return shortURL;
	}
}