package com.syntevo.plugin.removeemptydirectories;

import java.io.*;
import java.util.*;

import org.tmatesoft.svn.core.*;
import org.tmatesoft.svn.core.internal.wc17.*;
import org.tmatesoft.svn.core.internal.wc2.compat.*;
import org.tmatesoft.svn.core.wc.*;
import org.tmatesoft.svn.core.wc2.*;

import com.syntevo.openapi.log.*;
import com.syntevo.openapi.smartsvn.command.*;
import com.syntevo.openapi.smartsvn.file.*;
import com.syntevo.openapi.smartsvn.thread.*;

/**
 * @author syntevo GmbH
 */
final class REDSvnCommand implements ISvnCommand {

	// Constants ==============================================================

	private static final Logger LOG = LoggerFactory.createLogger(REDSvnCommand.class.getName());

	// Fields =================================================================

	private final SvnDirectory directory;

	// Setup ==================================================================

	public REDSvnCommand(SvnDirectory directory) {
		this.directory = directory;
	}

	// Implemented ============================================================

	public void perform(SvnClients svnClients, ProgressViewer progressViewer) throws SVNException {
		LOG.info("Removing empty directories within " + directory.getPath());

		final Map<File, List<File>> directoryToChildren = new HashMap<File, List<File>>();
		final Set<File> removableDirectories = new HashSet<File>();
		final SvnOperationFactory operationsFactory = svnClients.createOperationsFactory();
		fillDirectoryToChildren(directory, directoryToChildren, removableDirectories, operationsFactory);
		final List<File> emptyDirectories = determineEmptyDirectories(directoryToChildren, removableDirectories);
		removeEmptyDirectories(emptyDirectories, operationsFactory, progressViewer);
	}

	// Utils ==================================================================

	private static void fillDirectoryToChildren(SvnDirectory directory, final Map<File, List<File>> directoryToChildren, final Set<File> removableDirectories, SvnOperationFactory operationFactory) throws SVNException {
		final ISVNStatusHandler handler = new ISVNStatusHandler() {
			public void handleStatus(SVNStatus status) throws SVNException {
				final File file = status.getFile();
				if (status.getKind() == SVNNodeKind.DIR) {
					List<File> children = directoryToChildren.get(file);
					if (children == null) {
						children = new ArrayList<File>();
						directoryToChildren.put(file, children);
					}

					if (isRemovableDirectory(status)) {
						removableDirectories.add(file);
					}
				}

				final SVNStatusType contentsStatus = status.getContentsStatus();
				if (contentsStatus == SVNStatusType.STATUS_DELETED) {
					return;
				}

				final File parent = file.getParentFile();
				if (parent == null) {
					return;
				}

				List<File> children = directoryToChildren.get(parent);
				if (children == null) {
					children = new ArrayList<File>();
					directoryToChildren.put(parent, children);
				}

				children.add(file);
			}
		};

		final SvnGetStatus getStatus = operationFactory.createGetStatus();
		getStatus.setDepth(SVNDepth.INFINITY);
		getStatus.setRevision(SVNRevision.WORKING);
		getStatus.setReportAll(true);
		getStatus.setReportIgnored(true);
		getStatus.setSingleTarget(SvnTarget.fromFile(directory.getPath().getFile()));

		getStatus.setReceiver(new ISvnObjectReceiver<SvnStatus>() {
		    public void receive(SvnTarget target, SvnStatus status) throws SVNException {
		        final SVNWCContext context = getStatus.getOperationFactory().getWcContext();
		        handler.handleStatus(SvnCodec.status(context, status));
		    }
		});

		getStatus.run();

		getStatus.getRemoteRevision();
	}

	private static List<File> determineEmptyDirectories(Map<File, List<File>> directoryToChildren, Set<File> removableDirectories) {
		final List<File> emptyDirectories = new ArrayList<File>();
		for (boolean continu = true; continu;) {
			continu = false;

			for (File directory : new HashSet<File>(directoryToChildren.keySet())) {
				if (directoryToChildren.get(directory).size() > 0 || !removableDirectories.contains(directory)) {
					continue;
				}

				emptyDirectories.add(directory);
				directoryToChildren.remove(directory);
				final List<File> parentChildren = directoryToChildren.get(directory.getParentFile());
				if (parentChildren == null) {
					continue;
				}

				parentChildren.remove(directory);
				continu = true;
			}
		}
		return emptyDirectories;
	}

	private static void removeEmptyDirectories(List<File> emptyDirectories, SvnOperationFactory operationsFactory, ProgressViewer progressViewer) throws SVNException {
		// Reverse the empty directories, so we will process children before parents.
		Collections.sort(emptyDirectories, new Comparator<File>() {
			public int compare(File o1, File o2) {
				return o1.getAbsolutePath().compareTo(o2.getAbsolutePath());
			}
		});
		Collections.reverse(emptyDirectories);

		for (int index = 0; index < emptyDirectories.size(); index++) {
			final File directory = emptyDirectories.get(index);

			progressViewer.setProgress((double)index / (double)emptyDirectories.size());
			progressViewer.setMessage("Processing " + directory);

			final boolean[] removed = new boolean[1];

			final SvnGetStatus getStatus = operationsFactory.createGetStatus();
			getStatus.setDepth(SVNDepth.IMMEDIATES);
			getStatus.setRevision(SVNRevision.WORKING);
			getStatus.setReportAll(true);
			getStatus.setReportIgnored(true);
			getStatus.setSingleTarget(SvnTarget.fromFile(directory));
			getStatus.setReceiver(new ISvnObjectReceiver<SvnStatus>() {
			    public void receive(SvnTarget target, SvnStatus status) throws SVNException {
			        final SVNWCContext context = getStatus.getOperationFactory().getWcContext();
			        new ISVNStatusHandler() {
							public void handleStatus(final SVNStatus status) throws SVNException {
								final SVNStatusType contentsStatus = status.getContentsStatus();
								if (status.getFile().equals(directory)) {
									if (!isRemovableDirectory(status)) {
										throw new SVNException(SVNErrorMessage.create(SVNErrorCode.UNKNOWN, "Directory '" + directory + "' is modified."));
									}

									removed[0] = contentsStatus == SVNStatusType.STATUS_DELETED;
									return;
								}

								if (contentsStatus != SVNStatusType.STATUS_DELETED) {
									throw new SVNException(SVNErrorMessage.create(SVNErrorCode.UNKNOWN, "Directory '" + directory + "' is not empty now, but was before."));
								}
							}
						}.handleStatus(SvnCodec.status(context, status));
			    }
			});

			getStatus.run();

			getStatus.getRemoteRevision();

			if (removed[0]) {
				continue;
			}

			LOG.info("Scheduling empty directory " + directory + " for removal.");
			final SvnScheduleForRemoval remove = operationsFactory.createScheduleForRemoval();
			remove.setSingleTarget(SvnTarget.fromFile(directory));
			remove.setDeleteFiles(true);
			remove.run();
		}
	}

	private static boolean isRemovableDirectory(SVNStatus status) {
		if (status.getKind() != SVNNodeKind.DIR) {
			return false;
		}

		final SVNStatusType contentsStatus = status.getContentsStatus();
		final SVNStatusType properties = status.getPropertiesStatus();
		return contentsStatus == SVNStatusType.STATUS_DELETED || (contentsStatus == SVNStatusType.STATUS_NORMAL && (properties == SVNStatusType.STATUS_NONE || properties == SVNStatusType.STATUS_NORMAL));
	}
}
